/*
 *  TOPPERS/FMP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Advanced Standard Profile Kernel
 * 
 *  Copyright (C) 2006-2020 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 *  Copyright (C) 2017-2022 by TOPPERS PROJECT Educational Working Group.
 * 
 *  上記著作権者は，以下の(1)?(4)の条件を満たす場合に限り，本ソフトウェ
 *  ア（本ソフトウェアを改変したものを含む．以下同じ）を使用・複製・改
 *  変・再配布（以下，利用と呼ぶ）することを無償で許諾する．
 *  (1) 本ソフトウェアをソースコードの形で利用する場合には，上記の著作
 *      権表示，この利用条件および下記の無保証規定が，そのままの形でソー
 *      スコード中に含まれていること．
 *  (2) 本ソフトウェアを，ライブラリ形式など，他のソフトウェア開発に使
 *      用できる形で再配布する場合には，再配布に伴うドキュメント（利用
 *      者マニュアルなど）に，上記の著作権表示，この利用条件および下記
 *      の無保証規定を掲載すること．
 *  (3) 本ソフトウェアを，機器に組み込むなど，他のソフトウェア開発に使
 *      用できない形で再配布する場合には，次のいずれかの条件を満たすこ
 *      と．
 *    (a) 再配布に伴うドキュメント（利用者マニュアルなど）に，上記の著
 *        作権表示，この利用条件および下記の無保証規定を掲載すること．
 *    (b) 再配布の形態を，別に定める方法によって，TOPPERSプロジェクトに
 *        報告すること．
 *  (4) 本ソフトウェアの利用により直接的または間接的に生じるいかなる損
 *      害からも，上記著作権者およびTOPPERSプロジェクトを免責すること．
 *      また，本ソフトウェアのユーザまたはエンドユーザからのいかなる理
 *      由に基づく請求からも，上記著作権者およびTOPPERSプロジェクトを
 *      免責すること．
 * 
 *  本ソフトウェアは，無保証で提供されているものである．上記著作権者お
 *  よびTOPPERSプロジェクトは，本ソフトウェアに関して，特定の使用目的
 *  に対する適合性も含めて，いかなる保証も行わない．また，本ソフトウェ
 *  アの利用により直接的または間接的に生じたいかなる損害に関しても，そ
 *  の責任を負わない．
 * 
 *  @(#) $Id: chip_serial.c 698 2022-12-10 16:05:38Z roi $
 */

/*
 *  シリアルI/Oデバイス（SIO）ドライバ（rp2040用）
 */

#include <kernel.h>
#include <t_syslog.h>
#include "target_stddef.h"
#include "target_serial.h"
#include "target_syssvc.h"

/*
 *  SIL関数のマクロ定義
 */
#define sil_orw_mem(a, b)		sil_wrw_mem((a), sil_rew_mem(a) | (b))
#define sil_andw_mem(a, b)		sil_wrw_mem((a), sil_rew_mem(a) & ~(b))
#define sil_modw_mem(a, b, c)	sil_wrw_mem((a), (sil_rew_mem(a) & (~b)) | (c))

/*
 * レジスタ設定値
 */
#define PORT2SIOPID(x)	((x) + 1)
#define INDEX_PORT(x)	((x) - 1)
#define GET_SIOPCB(x)	(&siopcb_table[INDEX_PORT(x)])


#define REG_ALIAS_RW        0x00000000
#define REG_ALIAS_XOR       0x00001000
#define REG_ALIAS_SET       0x00002000
#define REG_ALIAS_CLR       0x00003000

/*
 *  IO BANK0
 */
#define TADR_IO_BANK0_BASE  0x40014000
#define TOFF_IO_BANK0_GPIO_CTRL        0x0004	/* GPIO control including function select and overrides */
  #define IO_BANK0_GPIO_CTRL_FUNCSEL_VALUE_XIP    0x00
  #define IO_BANK0_GPIO_CTRL_FUNCSEL_VALUE_SPI    0x01
  #define IO_BANK0_GPIO_CTRL_FUNCSEL_VALUE_UART   0x02
  #define IO_BANK0_GPIO_CTRL_FUNCSEL_VALUE_I2C    0x03
  #define IO_BANK0_GPIO_CTRL_FUNCSEL_VALUE_PWM    0x04
  #define IO_BANK0_GPIO_CTRL_FUNCSEL_VALUE_SIO    0x05
  #define IO_BANK0_GPIO_CTRL_FUNCSEL_VALUE_PIO0   0x06
  #define IO_BANK0_GPIO_CTRL_FUNCSEL_VALUE_PIO1   0x07
  #define IO_BANK0_GPIO_CTRL_FUNCSEL_VALUE_USB    0x09
  #define IO_BANK0_GPIO_CTRL_FUNCSEL_VALUE_NULL   0x1F
#define GPIO_SIZE                      8
#define NUM_BANK0_GPIO                 30

/*
 *  PADS BANK0
 */
#define TADR_PADS_BANK0_BASE 0x4001C000
#define TOFF_PADS_BANK0_GPIO           0x0004	/* (RW) Pad control register */
  #define PADS_BANK0_GPIO_SLEWFAST     0x00000001	/* SLEWFAST (0): Slew rate control */
  #define PADS_BANK0_GPIO_SCHMITT      0x00000002	/* SCHMITT (1): Enable schmitt trigger */
  #define PADS_BANK0_GPIO_PDE          0x00000004	/* PDE (1): Pull down enable */
  #define PADS_BANK0_GPIO_PUE          0x00000008	/* PUE (0): Pull up enable */
  #define PADS_BANK0_GPIO_DRIVE        0x00000030	/* DRIVE (1): Drive strength */
  #define PADS_BANK0_GPIO_IE           0x00000040	/* IE (1): Input enable */
  #define PADS_BANK0_GPIO_OD           0x00000080	/* OD (0): Output disable */

/*
 *  UNIVERSAL ASYNCHORONOUS RECEIVER TRANSMITTER
 */
#define TOFF_UART_UARTDR        0x0000		/* (RW) Data Register */
  #define UART_UARTDR_DATA        0x000000FF	/* DATA (0): Receive (read) data character */
  #define UART_UARTDR_FE          0x00000100	/* FE (0): Framing error */
  #define UART_UARTDR_PE          0x00000200	/* PE (0): Parity error */
  #define UART_UARTDR_BE          0x00000400	/* BE (0): Break error */
  #define UART_UARTDR_OE          0x00000800	/* OE (0): Overrun error */
#define TOFF_UART_UARTFR        0x0018		/* (R)  Flag Register */
  #define UART_UARTFR_CTS         0x00000001	/* CTS (0): Clear to send */
  #define UART_UARTFR_DSR         0x00000002	/* DSR (0): Data set ready */
  #define UART_UARTFR_DCD         0x00000004	/* DCD (0): Data carrier detect */
  #define UART_UARTFR_BUSY        0x00000008	/* BUSY (0): UART busy */
  #define UART_UARTFR_RXFE        0x00000010	/* RXFE (1): Receive FIFO empty */
  #define UART_UARTFR_TXFF        0x00000020	/* TXFF (0): Transmit FIFO full */
  #define UART_UARTFR_RXFF        0x00000040	/* RXFF (0): Receive FIFO full */
  #define UART_UARTFR_TXFE        0x00000080	/* TXFE (1): Transmit FIFO empty */
  #define UART_UARTFR_RI          0x00000100	/* RI (0): Ring indicator */
#define TOFF_UART_UARTIBRD      0x0024		/* (RW) Integer Baud Rate Register, */
#define TOFF_UART_UARTFBRD      0x0028		/* (RW) Fractional Baud Rate Register */
#define TOFF_UART_UARTLCR_H     0x002C		/* (RW) Line Control Register */
  #define UART_UARTLCR_H_BRK        0x00000001	/* BRK (0): Send break */
  #define UART_UARTLCR_H_PEN        0x00000002	/* PEN (0): Parity enable: 0 = parity is disabled and no parity bit added to the data frame 1 =... */
  #define UART_UARTLCR_H_EPS        0x00000004	/* EPS (0): Even parity select */
  #define UART_UARTLCR_H_STP2       0x00000008	/* STP2 (0): Two stop bits select */
  #define UART_UARTLCR_H_FEN        0x00000010	/* FEN (0): Enable FIFOs: 0 = FIFOs are disabled (character mode) that is, the FIFOs become... */
  #define UART_UARTLCR_H_WLEN       0x00000060	/* WLEN (0): Word length */
  #define UART_UARTLCR_H_SPS        0x00000080	/* SPS (0): Stick parity select */
#define TOFF_UART_UARTCR        0x0030	/* (RW) Control Register */
  #define UART_UARTCR_UARTEN        0x00000001	/* UARTEN (0): UART enable: 0 = UART is disabled */
  #define UART_UARTCR_SIREN         0x00000002	/* SIREN (0): SIR enable: 0 = IrDA SIR ENDEC is disabled */
  #define UART_UARTCR_SIRLP         0x00000004	/* SIRLP (0): SIR low-power IrDA mode */
  #define UART_UARTCR_LBE           0x00000080	/* LBE (0): Loopback enable */
  #define UART_UARTCR_TXE           0x00000100	/* TXE (1): Transmit enable */
  #define UART_UARTCR_RXE           0x00000200	/* RXE (1): Receive enable */
  #define UART_UARTCR_DTR           0x00000400	/* DTR (0): Data transmit ready */
  #define UART_UARTCR_RTS           0x00000800	/* RTS (0): Request to send */
  #define UART_UARTCR_OUT1          0x00001000	/* OUT1 (0): This bit is the complement of the UART Out1 (nUARTOut1) modem status output */
  #define UART_UARTCR_OUT2          0x00002000	/* OUT2 (0): This bit is the complement of the UART Out2 (nUARTOut2) modem status output */
  #define UART_UARTCR_RTSEN         0x00004000	/* RTSEN (0): RTS hardware flow control enable */
  #define UART_UARTCR_CTSEN         0x00008000	/* CTSEN (0): CTS hardware flow control enable */
#define TOFF_UART_UARTIFLS      0x0034	/* (RW) Interrupt FIFO Level Select Register */
  #define UART_UARTIFLS_TXIFLSEL    0x00000007	/* TXIFLSEL (0x2): Transmit interrupt FIFO level select */
  #define UART_UARTIFLS_RXIFLSEL    0x00000038	/* RXIFLSEL (0x2): Receive interrupt FIFO level select */
#define TOFF_UART_UARTIMSC      0x0038	/* (RW) Interrupt Mask Set/Clear Register */
  #define UART_UARTIMSC_RIMIM       0x00000001	/* RIMIM (0): nUARTRI modem interrupt mask */
  #define UART_UARTIMSC_CTSMIM      0x00000002	/* CTSMIM (0): nUARTCTS modem interrupt mask */
  #define UART_UARTIMSC_DCDMIM      0x00000004	/* DCDMIM (0): nUARTDCD modem interrupt mask */
  #define UART_UARTIMSC_DSRMIM      0x00000008	/* DSRMIM (0): nUARTDSR modem interrupt mask */
  #define UART_UARTIMSC_RXIM        0x00000010	/* RXIM (0): Receive interrupt mask */
  #define UART_UARTIMSC_TXIM        0x00000020	/* TXIM (0): Transmit interrupt mask */
  #define UART_UARTIMSC_RTIM        0x00000040	/* RTIM (0): Receive timeout interrupt mask */
  #define UART_UARTIMSC_FEIM        0x00000080	/* FEIM (0): Framing error interrupt mask */
  #define UART_UARTIMSC_PEIM        0x00000100	/* PEIM (0): Parity error interrupt mask */
  #define UART_UARTIMSC_BEIM        0x00000200	/* BEIM (0): Break error interrupt mask */
  #define UART_UARTIMSC_OEIM        0x00000400	/* OEIM (0): Overrun error interrupt mask */
#define TOFF_UART_UARTRIS       0x003C	/* (R)  Raw Interrupt Status Register */
  #define UART_UARTRIS_RIRMIS       0x00000001	/* RIRMIS (0): nUARTRI modem interrupt status */
  #define UART_UARTRIS_CTSRMIS      0x00000002	/* CTSRMIS (0): nUARTCTS modem interrupt status */
  #define UART_UARTRIS_DCDRMIS      0x00000004	/* DCDRMIS (0): nUARTDCD modem interrupt status */
  #define UART_UARTRIS_DSRRMIS      0x00000008	/* DSRRMIS (0): nUARTDSR modem interrupt status */
  #define UART_UARTRIS_RXRIS        0x00000010	/* RXRIS (0): Receive interrupt status */
  #define UART_UARTRIS_TXRIS        0x00000020	/* TXRIS (0): Transmit interrupt status */
  #define UART_UARTRIS_RTRIS        0x00000040	/* RTRIS (0): Receive timeout interrupt status */
  #define UART_UARTRIS_FERIS        0x00000080	/* FERIS (0): Framing error interrupt status */
  #define UART_UARTRIS_PERIS        0x00000100	/* PERIS (0): Parity error interrupt status */
  #define UART_UARTRIS_BERIS        0x00000200	/* BERIS (0): Break error interrupt status */
  #define UART_UARTRIS_OERIS        0x00000400	/* OERIS (0): Overrun error interrupt status */
#define TOFF_UART_UARTDMACR     0x0048	/* (RW) DMA Control Register */
  #define UART_UARTDMACR_RXDMAE     0x00000001	/* RXDMAE (0): Receive DMA enable */
  #define UART_UARTDMACR_TXDMAE     0x00000002	/* TXDMAE (0): Transmit DMA enable */
  #define UART_UARTDMACR_DMAONERR   0x00000004	/* DMAONERR (0): DMA on error */

#define UART_MODE_MASK	(UART_UARTLCR_H_WLEN | UART_UARTLCR_H_STP2 | UART_UARTLCR_H_PEN | UART_UARTLCR_H_EPS)
#define UART_MODE_VALUE ((8-5) << 5)	/* data 8bits */

/*
 *  シリアルI/Oポート初期化ブロックの定義
 */
typedef struct sio_port_initialization_block {
	uint32_t base;
	INTNO    intno_usart;
	uint32_t reset_addr;
	uint32_t reset_done_addr;
	uint32_t reset;
} SIOPINIB;

/*
 *  兼用GPIOポート初期化ブロックの定義
 */
typedef struct gpio_port_initialization_block {
	uint32_t txpin;
	uint32_t rxpin;
} GPIOINIB;

/*
 *  シリアルI/Oポート管理ブロックの定義
 */
struct sio_port_control_block {
	const SIOPINIB  *p_siopinib;  /* シリアルI/Oポート初期化ブロック */
	const GPIOINIB  *p_gpioinib;  /* 兼用GPIOポート初期化ブロック */
	intptr_t        exinf;        /* 拡張情報 */
	bool_t          opnflg;       /* オープン済みフラグ */
};

/*
 * シリアルI/Oポート初期化ブロック
 */
const SIOPINIB siopinib_table[TNUM_SIOP] = {
	{(uint32_t)USART1_BASE,         (INTNO)INTNO_SIO1,
	 USART1_RESET_ADDR, USART1_RESET_CHECK, RESET_U1_BIT},
#if TNUM_SIOP >= 2
	{(uint32_t)USART2_BASE,         (INTNO)INTNO_SIO2,
	 USART2_RESET_ADDR, USART2_RESET_CHECK, RESET_U2_BIT},
#endif
};

/*
 * 兼用GPIOポート初期化ブロック
 */
const GPIOINIB gpioinib_table[TNUM_SIOP] = {
	{(uint32_t)TX1_PINPOS,
     (uint32_t)RX1_PINPOS
#if TNUM_SIOP >= 2
    },
	{(uint32_t)TX2_PINPOS,
     (uint32_t)RX2_PINPOS
#endif
    }
};

/*
 *  シリアルI/Oポート管理ブロックのエリア
 */
SIOPCB	siopcb_table[TNUM_SIOP];

/*
 *  シリアルI/OポートIDから管理ブロックを取り出すためのマクロ
 */
#define INDEX_SIOP(siopid)	((uint_t)((siopid) - 1))
#define get_siopcb(siopid)	(&(siopcb_table[INDEX_SIOP(siopid)]))

/*
 *  ピンMUX設定
 */
static void
pinmux_setup(uint8_t pin, uint8_t modefunc)
{
	uint32_t pbase, gbase, mask;

	if(pin >= NUM_BANK0_GPIO)
		return;
	if(modefunc > 0x1F)
		return;
	pbase = TADR_PADS_BANK0_BASE+TOFF_PADS_BANK0_GPIO+pin*4;
	gbase = TADR_IO_BANK0_BASE+pin*GPIO_SIZE;

	mask = ((sil_rew_mem((uint32_t *)(pbase)) ^ PADS_BANK0_GPIO_IE) & (PADS_BANK0_GPIO_IE | PADS_BANK0_GPIO_OD));
	sil_wrw_mem((uint32_t *)(pbase+REG_ALIAS_XOR), mask);

	sil_wrw_mem((uint32_t *)(gbase+TOFF_IO_BANK0_GPIO_CTRL), modefunc);
}

volatile void
put_hex(char a, int val)
{
	int i, j;
	target_fput_log(a);
	target_fput_log(' ');
	for(i = 28 ; i >= 0 ; i-= 4){
		j = (val >> i) & 0xf;;
		if(j > 9)
			j += ('A'-10);
		else
			j += '0';
		target_fput_log(j);
	}
	target_fput_log('\n');
}

/*
 *  SIOドライバの初期化
 */
void
sio_initialize(intptr_t exinf)
{
	SIOPCB	*p_siopcb;
	uint_t	i;

	/*
	 *  シリアルI/Oポート管理ブロックの初期化
	 */
	for (p_siopcb = siopcb_table, i = 0; i < TNUM_SIOP; p_siopcb++, i++) {
		p_siopcb->p_siopinib = &(siopinib_table[i]);
		p_siopcb->p_gpioinib = &(gpioinib_table[i]);
		p_siopcb->opnflg = false;
	}
}


/*
 *  シリアルI/Oポートのオープン
 */
SIOPCB *
sio_opn_por(ID siopid, intptr_t exinf)
{
	SIOPCB          *p_siopcb;
	const SIOPINIB  *p_siopinib;
	const GPIOINIB  *p_gpioinib;
	bool_t   opnflg;
	ER       ercd;
	uint32_t base, reset, clock, mask;
	uint32_t baud_rate_div, baud_ibrd, baud_fbrd;

	p_siopcb = get_siopcb(siopid);
	p_siopinib = p_siopcb->p_siopinib;
	p_gpioinib = p_siopcb->p_gpioinib;


	/*
	 *  オープンしたポートがあるかをopnflgに読んでおく．
	 */
	opnflg = p_siopcb->opnflg;

	p_siopcb->exinf = exinf;
	base  = p_siopinib->base;
	reset = p_siopinib->reset;

	if((clock = get_Clock(CLK_PERI)) == 0)
		return NULL;

	/*
	 *  ハードウェアの初期化
	 */
	pinmux_setup(p_gpioinib->txpin, IO_BANK0_GPIO_CTRL_FUNCSEL_VALUE_UART);
	pinmux_setup(p_gpioinib->rxpin, IO_BANK0_GPIO_CTRL_FUNCSEL_VALUE_UART);

	/*
	 *  UARTをリセット
	 */
	sil_wrw_mem((uint32_t *)(p_siopinib->reset_addr+REG_ALIAS_SET), reset);
	sil_wrw_mem((uint32_t *)(p_siopinib->reset_addr+REG_ALIAS_CLR), reset);
	while((sil_rew_mem((uint32_t *)(p_siopinib->reset_done_addr)) & reset) == 0);

	/*
	 *  BAUD設定
	 */
	baud_rate_div = (8 * clock / BPS_SETTING);
    baud_ibrd = baud_rate_div >> 7;
	if(baud_ibrd == 0){
		baud_ibrd = 1;
		baud_fbrd = 0;
    }
	else if(baud_ibrd >= 65535){
		baud_ibrd = 65535;
		baud_fbrd = 0;
	}
	else{
		baud_fbrd = ((baud_rate_div & 0x7f) + 1) / 2;
	}

	/*
	 *  Baudrate Dividor設定
	 */
	sil_wrw_mem((uint32_t *)(base+TOFF_UART_UARTIBRD), baud_ibrd);
	sil_wrw_mem((uint32_t *)(base+TOFF_UART_UARTFBRD), baud_fbrd);

	/*
	 *  ラインコントロールレジスタにダミー設定
	 */
	sil_wrw_mem((uint32_t *)(base+TOFF_UART_UARTLCR_H+REG_ALIAS_SET), 0);

	/*
	 *  通信モード設定
	 */
	sil_modw_mem((uint32_t *)(base+TOFF_UART_UARTLCR_H), UART_MODE_MASK, UART_MODE_VALUE);

	/*
	 *  送受信有効化
	 */
	sil_wrw_mem((uint32_t *)(base+TOFF_UART_UARTCR), (UART_UARTCR_UARTEN | UART_UARTCR_TXE | UART_UARTCR_RXE));

	/*
	 *  RX/TX DMA設定
	 */
	sil_wrw_mem((uint32_t *)(base+TOFF_UART_UARTDMACR), (UART_UARTDMACR_TXDMAE | UART_UARTDMACR_RXDMAE));

	/*
	 *  FIFO無効化
	 */
	sil_andw_mem((uint32_t *)(base+TOFF_UART_UARTLCR_H), UART_UARTLCR_H_FEN);

	/*
	 *  割込み設定
	 */
	sil_wrw_mem((uint32_t *)(base+TOFF_UART_UARTIMSC), UART_UARTIMSC_RTIM);
	mask = (sil_rew_mem((uint32_t *)(base+TOFF_UART_UARTIFLS)) & (UART_UARTIFLS_TXIFLSEL | UART_UARTIFLS_RXIFLSEL));
	sil_wrw_mem((uint32_t *)(base+TOFF_UART_UARTIFLS+REG_ALIAS_XOR), mask);

	p_siopcb->opnflg = true;
	sil_dly_nse(1000*1000);

	/*
	 *  シリアルI/O割込みのマスクを解除する．
	 */
	if (!opnflg) {
		ercd = ena_int(p_siopinib->intno_usart);
		assert(ercd == E_OK);
	}

	return(p_siopcb);
}

/*
 *  シリアルI/Oポートのクローズ
 */
void
sio_cls_por(SIOPCB *p_siopcb)
{
	const SIOPINIB  *p_siopinib;

	p_siopinib = p_siopcb->p_siopinib;

	/*
	 *  UARTをリセット
	 */
	sil_wrw_mem((uint32_t *)(p_siopinib->reset_addr+REG_ALIAS_SET), p_siopinib->reset);

	/*
	 *  シリアルI/O割込みをマスクする．
	 */
	if ((p_siopcb->opnflg)) {
		dis_int(p_siopcb->p_siopinib->intno_usart);
	}
	p_siopcb->opnflg = false;
}

/*
 *  SIOの割込みサービスルーチン
 */

Inline bool_t
sio_putiready(SIOPCB* p_siopcb)
{
	const SIOPINIB  *p_siopinib = p_siopcb->p_siopinib;
	uint32_t base = p_siopinib->base;
	uint32_t imsc = sil_rew_mem((uint32_t *)(base+TOFF_UART_UARTIMSC));
	uint32_t ris  = sil_rew_mem((uint32_t *)(base+TOFF_UART_UARTRIS));

	if((imsc & UART_UARTIMSC_TXIM) != 0 && (ris & UART_UARTRIS_TXRIS) != 0){
		return 1;
	}
	return 0;
}

Inline bool_t
sio_getiready(SIOPCB* p_siopcb)
{
	const SIOPINIB  *p_siopinib = p_siopcb->p_siopinib;
	uint32_t base = p_siopinib->base;
	uint32_t imsc = sil_rew_mem((uint32_t *)(base+TOFF_UART_UARTIMSC));
	uint32_t ufr  = sil_rew_mem((uint32_t *)(base+TOFF_UART_UARTFR));

	if ((imsc & UART_UARTIMSC_RXIM) != 0 && (ufr & UART_UARTFR_RXFE) == 0){
		return 1;
	}
	return 0;
}

void
sio_usart_isr(intptr_t exinf)
{
	SIOPCB          *p_siopcb;

	p_siopcb = get_siopcb(exinf);

	if (sio_getiready(p_siopcb)) {
		sio_irdy_rcv(p_siopcb->exinf);
	}
	if (sio_putiready(p_siopcb)) {
		sio_irdy_snd(p_siopcb->exinf);
	}
}

Inline bool_t
sio_putready(SIOPCB* p_siopcb)
{
	const SIOPINIB  *p_siopinib = p_siopcb->p_siopinib;
	if((sil_rew_mem((uint32_t *)(p_siopinib->base+TOFF_UART_UARTFR)) & UART_UARTFR_TXFF) == 0)
		return true;
	else
		return false;
}

/*
 *  シリアルI/Oポートへの文字送信
 */
bool_t
sio_snd_chr(SIOPCB *p_siopcb, char c)
{
	if (sio_putready(p_siopcb)) {
		sil_wrw_mem((uint32_t *)(p_siopcb->p_siopinib->base+TOFF_UART_UARTDR), (uint32_t)c);
		return true;
	}
	return false;
}

/*
 *  シリアルI/Oポートからの文字受信
 */
int_t
sio_rcv_chr(SIOPCB *p_siopcb)
{
	return sil_rew_mem((uint32_t *)(p_siopcb->p_siopinib->base+TOFF_UART_UARTDR)) & 0xFF;
}

/*
 *  シリアルI/Oポートからのコールバックの許可
 */
void
sio_ena_cbr(SIOPCB *p_siopcb, uint_t cbrtn)
{
	const SIOPINIB  *p_siopinib = p_siopcb->p_siopinib;
	uint32_t base = p_siopinib->base;

	switch (cbrtn) {
	case SIO_RDY_SND:
		sil_orw_mem((uint32_t *)(base+TOFF_UART_UARTIMSC), UART_UARTIMSC_TXIM);
		break;
	case SIO_RDY_RCV:
		sil_orw_mem((uint32_t *)(base+TOFF_UART_UARTIMSC), UART_UARTIMSC_RXIM);
		break;
	}
}

/*
 *  シリアルI/Oポートからのコールバックの禁止
 */
void
sio_dis_cbr(SIOPCB *p_siopcb, uint_t cbrtn)
{
	const SIOPINIB  *p_siopinib = p_siopcb->p_siopinib;
	uint32_t base = p_siopinib->base;

	switch (cbrtn) {
	case SIO_RDY_SND:
		sil_andw_mem((uint32_t *)(base+TOFF_UART_UARTIMSC), UART_UARTIMSC_TXIM);
		break;
	case SIO_RDY_RCV:
		sil_andw_mem((uint32_t *)(base+TOFF_UART_UARTIMSC), UART_UARTIMSC_RXIM);
		break;
	}
}

/*
 *  1文字出力（ポーリングでの出力）
 */
void sio_pol_snd_chr(int8_t c, ID siopid)
{
	const SIOPINIB  *p_siopinib = &siopinib_table[INDEX_PORT(siopid)];
	uint32_t base = p_siopinib->base;

	sil_wrw_mem((uint32_t *)(base+TOFF_UART_UARTDR), (uint32_t)c);
	while ((sil_rew_mem((uint32_t *)(base+TOFF_UART_UARTFR)) & UART_UARTFR_TXFF) != 0);

	/*
	 *  出力が完全に終わるまで待つ
	 */
	volatile int n = SystemFrequency/BPS_SETTING;
	while(n--);
}

/*
 *  ターゲットのシリアル初期化
 */
void chip_uart_init(ID siopid)
{
	const SIOPINIB	*p_siopinib = &siopinib_table[INDEX_PORT(siopid)];
	const GPIOINIB  *p_gpioinib = &gpioinib_table[INDEX_PORT(siopid)];
	uint32_t base, reset, clock, mask;
	uint32_t baud_rate_div, baud_ibrd, baud_fbrd;

	base = p_siopinib->base;
	reset = p_siopinib->reset;

	if((clock = get_Clock(CLK_PERI)) == 0)
		return;

	/*
	 *  ハードウェアの初期化
	 */
	pinmux_setup(p_gpioinib->txpin, IO_BANK0_GPIO_CTRL_FUNCSEL_VALUE_UART);
	pinmux_setup(p_gpioinib->rxpin, IO_BANK0_GPIO_CTRL_FUNCSEL_VALUE_UART);

	/*
	 *  UARTをリセット
	 */
	sil_wrw_mem((uint32_t *)(p_siopinib->reset_addr+REG_ALIAS_SET), reset);
	sil_wrw_mem((uint32_t *)(p_siopinib->reset_addr+REG_ALIAS_CLR), reset);
	while((sil_rew_mem((uint32_t *)(p_siopinib->reset_done_addr)) & reset) == 0);

	/*
	 *  BAUD設定
	 */
	baud_rate_div = (8 * clock / BPS_SETTING);
    baud_ibrd = baud_rate_div >> 7;
	if(baud_ibrd == 0){
		baud_ibrd = 1;
		baud_fbrd = 0;
    }
	else if(baud_ibrd >= 65535){
		baud_ibrd = 65535;
		baud_fbrd = 0;
	}
	else{
		baud_fbrd = ((baud_rate_div & 0x7f) + 1) / 2;
	}

	/*
	 *  Baudrate Dividor設定
	 */
	sil_wrw_mem((uint32_t *)(base+TOFF_UART_UARTIBRD), baud_ibrd);
	sil_wrw_mem((uint32_t *)(base+TOFF_UART_UARTFBRD), baud_fbrd);

	/*
	 *  ラインコントロールレジスタにダミー設定
	 */
	sil_wrw_mem((uint32_t *)(base+TOFF_UART_UARTLCR_H+REG_ALIAS_SET), 0);

	/*
	 *  通信モード設定
	 */
	sil_modw_mem((uint32_t *)(base+TOFF_UART_UARTLCR_H), UART_MODE_MASK, UART_MODE_VALUE);

	/*
	 *  送受信有効化
	 */
	sil_wrw_mem((uint32_t *)(base+TOFF_UART_UARTCR), (UART_UARTCR_UARTEN | UART_UARTCR_TXE | UART_UARTCR_RXE));

	/*
	 *  RX/TX DMA設定
	 */
	sil_wrw_mem((uint32_t *)(base+TOFF_UART_UARTDMACR), (UART_UARTDMACR_TXDMAE | UART_UARTDMACR_RXDMAE));

	/*
	 *  FIFO無効化
	 */
	sil_andw_mem((uint32_t *)(base+TOFF_UART_UARTLCR_H), UART_UARTLCR_H_FEN);

	/*
	 *  割込み設定
	 */
	mask = (sil_rew_mem((uint32_t *)(base+TOFF_UART_UARTIFLS)) & (UART_UARTIFLS_TXIFLSEL | UART_UARTIFLS_RXIFLSEL));
	sil_wrw_mem((uint32_t *)(base+TOFF_UART_UARTIFLS+REG_ALIAS_XOR), mask);
}
