/*
 *  TOPPERS/FMP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Advanced Standard Profile Kernel
 * 
 *  Copyright (C) 2008-2011 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 *  Copyright (C) 2016-2022 by TOPPERS PROJECT Educational Working Group.
 * 
 *  上記著作権者は，以下の(1)～(4)の条件を満たす場合に限り，本ソフトウェ
 *  ア（本ソフトウェアを改変したものを含む．以下同じ）を使用・複製・改
 *  変・再配布（以下，利用と呼ぶ）することを無償で許諾する．
 *  (1) 本ソフトウェアをソースコードの形で利用する場合には，上記の著作
 *      権表示，この利用条件および下記の無保証規定が，そのままの形でソー
 *      スコード中に含まれていること．
 *  (2) 本ソフトウェアを，ライブラリ形式など，他のソフトウェア開発に使
 *      用できる形で再配布する場合には，再配布に伴うドキュメント（利用
 *      者マニュアルなど）に，上記の著作権表示，この利用条件および下記
 *      の無保証規定を掲載すること．
 *  (3) 本ソフトウェアを，機器に組み込むなど，他のソフトウェア開発に使
 *      用できない形で再配布する場合には，次のいずれかの条件を満たすこ
 *      と．
 *    (a) 再配布に伴うドキュメント（利用者マニュアルなど）に，上記の著
 *        作権表示，この利用条件および下記の無保証規定を掲載すること．
 *    (b) 再配布の形態を，別に定める方法によって，TOPPERSプロジェクトに
 *        報告すること．
 *  (4) 本ソフトウェアの利用により直接的または間接的に生じるいかなる損
 *      害からも，上記著作権者およびTOPPERSプロジェクトを免責すること．
 *      また，本ソフトウェアのユーザまたはエンドユーザからのいかなる理
 *      由に基づく請求からも，上記著作権者およびTOPPERSプロジェクトを
 *      免責すること．
 * 
 *  本ソフトウェアは，無保証で提供されているものである．上記著作権者お
 *  よびTOPPERSプロジェクトは，本ソフトウェアに関して，特定の使用目的
 *  に対する適合性も含めて，いかなる保証も行わない．また，本ソフトウェ
 *  アの利用により直接的または間接的に生じたいかなる損害に関しても，そ
 *  の責任を負わない．
 * 
 *  @(#) $Id: chip_timer.h 292 698 2022-11-10 15:58:05Z roi $
 */

/*
 *  タイマドライバ（PRC2TIMER用）
 */
#include <sil.h>
#include "kernel_impl.h"
#include "time_event.h"
#include "target_timer.h"

#define TADR_TIMER_BASE     0x40054000
#define TOFF_TIMER_ALARM0     0x0010	/* (RW) Arm alarm 0, and configure the time it will fire. */
#define TOFF_TIMER_TIMERAWL   0x0028	/* (R)  Raw read from bits 31:0 of time (no side effects) */
#define TOFF_TIMER_INTR       0x0034	/* (W)  Raw Interrupts */
#define TOFF_TIMER_INTE       0x0038	/* (RW) Interrupt Enable */
#define TOFF_TIMER_INTS       0x0040	/* (R) Interrupt status after masking & forcing  */

/*
 *  タイマの起動処理
 */
void
prc2_timer_initialize(intptr_t exinf)
{
	uint64_t delay_time;
	uint32_t inte = sil_rew_mem((uint32_t *)(TADR_TIMER_BASE+TOFF_TIMER_INTE));

	inte |= (1<<PRC2_TIMERNO);
	sil_wrw_mem((uint32_t *)(TADR_TIMER_BASE+TOFF_TIMER_INTE), inte);
	delay_time = sil_rew_mem((uint32_t *)(TADR_TIMER_BASE+TOFF_TIMER_TIMERAWL)) + WRAP_TIME;
	sil_wrw_mem((uint32_t *)(TADR_TIMER_BASE+TOFF_TIMER_ALARM0+PRC2_TIMERNO*4), (uint32_t)delay_time);
}

/*
 *  タイマの停止処理
 */
void
prc2_timer_terminate(intptr_t exinf)
{
	uint32_t inte = sil_rew_mem((uint32_t *)(TADR_TIMER_BASE+TOFF_TIMER_INTE));

	inte &= ~(1<<PRC2_TIMERNO);
	sil_wrw_mem((uint32_t *)(TADR_TIMER_BASE+TOFF_TIMER_INTE), inte);
}

/*
 *  タイマ割込みハンドラ
 */
void
prc2_timer_handler(void)
{
	uint64_t delay_time;
	uint32_t ints = sil_rew_mem((uint32_t *)(TADR_TIMER_BASE+TOFF_TIMER_INTS));
	/*
	 *  割込みクリア
	 */
	sil_wrw_mem((uint32_t *)(TADR_TIMER_BASE+TOFF_TIMER_INTR), ints);

	/*
	 *  リロード時間設定
	 */
	delay_time = sil_rew_mem((uint32_t *)(TADR_TIMER_BASE+TOFF_TIMER_TIMERAWL)) + WRAP_TIME;
	sil_wrw_mem((uint32_t *)(TADR_TIMER_BASE+TOFF_TIMER_ALARM0+PRC2_TIMERNO*4), (uint32_t)delay_time);

	i_begin_int(0x20000+INTNO_STIMER);
	signal_time();                    /* タイムティックの供給 */
	i_end_int(0x20000+INTNO_STIMER);
}
