/*
 *  TOPPERS/FMP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Advanced Standard Profile Kernel
 * 
 *  Copyright (C) 2007,2011-2015 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 *  Copyright (C) 2017-2023 by TOPPERS PROJECT Educational Working Group.
 * 
 *  上記著作権者は，以下の(1)～(4)の条件を満たす場合に限り，本ソフトウェ
 *  ア（本ソフトウェアを改変したものを含む．以下同じ）を使用・複製・改
 *  変・再配布（以下，利用と呼ぶ）することを無償で許諾する．
 *  (1) 本ソフトウェアをソースコードの形で利用する場合には，上記の著作
 *      権表示，この利用条件および下記の無保証規定が，そのままの形でソー
 *      スコード中に含まれていること．
 *  (2) 本ソフトウェアを，ライブラリ形式など，他のソフトウェア開発に使
 *      用できる形で再配布する場合には，再配布に伴うドキュメント（利用
 *      者マニュアルなど）に，上記の著作権表示，この利用条件および下記
 *      の無保証規定を掲載すること．
 *  (3) 本ソフトウェアを，機器に組み込むなど，他のソフトウェア開発に使
 *      用できない形で再配布する場合には，次のいずれかの条件を満たすこ
 *      と．
 *    (a) 再配布に伴うドキュメント（利用者マニュアルなど）に，上記の著
 *        作権表示，この利用条件および下記の無保証規定を掲載すること．
 *    (b) 再配布の形態を，別に定める方法によって，TOPPERSプロジェクトに
 *        報告すること．
 *  (4) 本ソフトウェアの利用により直接的または間接的に生じるいかなる損
 *      害からも，上記著作権者およびTOPPERSプロジェクトを免責すること．
 *      また，本ソフトウェアのユーザまたはエンドユーザからのいかなる理
 *      由に基づく請求からも，上記著作権者およびTOPPERSプロジェクトを
 *      免責すること．
 * 
 *  本ソフトウェアは，無保証で提供されているものである．上記著作権者お
 *  よびTOPPERSプロジェクトは，本ソフトウェアに関して，特定の使用目的
 *  に対する適合性も含めて，いかなる保証も行わない．また，本ソフトウェ
 *  アの利用により直接的または間接的に生じたいかなる損害に関しても，そ
 *  の責任を負わない．
 * 
 *  @(#) $Id: target_syssvc.h 698 2023-02-14 11:31:05Z roi $
 */

/*
 *  システムサービスのターゲット依存部（raspberrypi_pico用）
 *
 *  システムサービスのターゲット依存部のインクルードファイル．このファ
 *  イルの内容は，コンポーネント記述ファイルに記述され，このファイルは
 *  無くなる見込み．
 */

#ifndef TOPPERS_TARGET_SYSSVC_H
#define TOPPERS_TARGET_SYSSVC_H

/*
 *  ターゲットシステムのハードウェア資源の定義
 */
#include <sil.h>
#include <target_serial.h>

/*
 *  クロックの定義
 */
#define CLK_GPOUT0          0	/* GPIO Muxing 0 */
#define CLK_GPOUT1          1	/* GPIO Muxing 1 */
#define CLK_GPOUT2          2	/* GPIO Muxing 2 */
#define CLK_GPOUT3          3	/* GPIO Muxing 3 */
#define CLK_REF             4	/* Watchdog and timers reference clock */
#define CLK_SYS             5	/* Processors, bus fabric, memory, memory mapped registers */
#define CLK_PERI			6	/* Peripheral clock for UART and SPI */
#define CLK_USB             7	/* USB clock */
#define CLK_ADC             8	/* ADC clock */
#define CLK_RTC				9	/* Real time clock */
#define NUM_CLK             10

/*
 *  コア参照アドレスの定義
 */
#define TADR_SIO_CPUID      0xD0000000


#ifndef TOPPERS_MACRO_ONLY

/*
 *  システムクロックの定義
 */
extern  uint32_t SystemFrequency;
#define SysFreHCLK          SystemFrequency	/* HCLK = MasterClock / 1 */
#define SysFrePCLK1         SystemFrequency	/* PCLK1 = HCLK / 1 */

#define SYS_CLOCK		    (SystemFrequency)

#endif /* TOPPERS_MACRO_ONLY */

/*
 *  トレースログに関する設定
 */
#ifdef TOPPERS_ENABLE_TRACE
#include "logtrace/trace_config.h"
#endif /* TOPPERS_ENABLE_TRACE */

/*
 *  起動メッセージのターゲットシステム名
 */
#define TARGET_NAME    "Rasperry Pi Pico(Cortex-M0+)"

/*
 *  起動メッセージの著作権表示
 */
#define TARGET_COPYRIGHT \
"Copyright (C) 2016-2023 by Education Working Group TOPPERS PROJECT\n" \

#ifndef TOPPERS_MACRO_ONLY

typedef void (*func_t)(void);

/*
 *  システムログの低レベル出力のための文字出力
 *
 *  ターゲット依存の方法で，文字cを表示/出力/保存する．
 */
extern void	target_fput_log(int8_t c);

/*
 *  クロックを取得する
 */
extern uint32_t get_Clock(uint8_t index);

/*
 *  クロックテーブルを再設定する
 */
extern void recover_ClocksValue(void);

/*
 *  実行コアを取得する
 */
extern uint32_t get_CoreNo(void);


/*
 *  コア1を起動する
 */
extern void mprc_initialize(func_t entry, uint32_t *sp, uint32_t vec);

/*
 *  IPIのイベント送信を行う
 */
extern ER send_ipi_event(uint32_t event);

/*
 *  IPIのイベント受信コールバック関数を設定する
 */
extern void set_ipi_callback(void *func);

#endif /* TOPPERS_MACRO_ONLY */

/*
 *  サポートするシリアルポートの数
 */
#define TNUM_PORT        TNUM_SIOP

/*
 *  ログタスクが使用するポートID
 */
#define LOGTASK_PORTID   SIO_PORTID

/*
 *  ボーレート
 */
#define BPS_SETTING		(115200)

/*
 *  システムログタスク関連の定数の定義
 *
 *  デフォルト値の通り．
 */

#endif /* TOPPERS_TARGET_SYSSVC_H */
